"""
PrimeLunch Utils
"""

from __future__ import annotations

import datetime
import html
import logging
import os
import pathlib
from typing import Any, Optional, Union

import click
import numpy as np
import pandas as pd
from numpy import datetime64
from rich import print, table
from rich.prompt import Confirm

from lunchable import TransactionUpdateObject
from lunchable._config.logging_config import set_up_logging
from lunchable.models import CategoriesObject, TransactionObject, UserObject
from lunchable.plugins.base.pandas_app import LunchablePandasApp

logger = logging.getLogger(__name__)


class PrimeLunch(LunchablePandasApp):
    """
    PrimeLunch: Amazon Notes Updater
    """

    def __init__(
        self,
        file_path: Union[str, os.PathLike[str], pathlib.Path],
        time_window: int = 7,
        access_token: Optional[str] = None,
    ) -> None:
        """
        Initialize and set internal data
        """
        super().__init__(cache_time=0, access_token=access_token)
        self.file_path = pathlib.Path(file_path)
        self.time_window = time_window

    def amazon_to_df(self) -> pd.DataFrame:
        """
        Read an Amazon Data File to a DataFrame

        This is pretty simple, except duplicate header rows need to be cleaned

        Returns
        -------
        pd.DataFrame
        """
        dt64: np.dtype[datetime64] = np.dtype("datetime64[ns]")
        expected_columns = {
            "order id": str,
            "items": str,
            "to": str,
            "date": dt64,
            "total": np.float64,
            "shipping": np.float64,
            "gift": np.float64,
            "refund": np.float64,
            "payments": str,
        }
        amazon_df = pd.read_csv(
            self.file_path,
            usecols=expected_columns.keys(),
        )
        header_row_eval = pd.concat(
            [amazon_df[item] == item for item in expected_columns.keys()], axis=1
        ).all(axis=1)
        duplicate_header_rows = np.where(header_row_eval)[0]
        amazon_df.drop(duplicate_header_rows, axis=0, inplace=True)
        amazon_df["total"] = (
            amazon_df["total"].astype("string").str.replace(",", "").astype(np.float64)
        )
        amazon_df = amazon_df.astype(dtype=expected_columns, copy=True, errors="raise")
        logger.info("Amazon Data File loaded: %s", self.file_path)
        return amazon_df

    @classmethod
    def filter_amazon_transactions(cls, df: pd.DataFrame) -> pd.DataFrame:
        """
        Filter a DataFrame to Amazon Transactions

        Parameters
        ----------
        df: pd.DataFrame

        Returns
        -------
        pd.DataFrame
        """
        amazon_transactions = df.copy()
        amazon_transactions["original_name"] = amazon_transactions[
            "original_name"
        ].fillna("")
        amazon_transactions = amazon_transactions[
            amazon_transactions.payee.str.match(
                r"(?i)(Amazon|AMZN|Whole Foods)(\s?(Prime|Marketplace|MKTP)|\.\w+)?",
                case=False,
            )
            | amazon_transactions.original_name.str.match(
                r"(?i)(Amazon|AMZN|Whole Foods)(\s?(Prime|Marketplace|MKTP)|\.\w+)?",
                case=False,
            )
        ]
        return amazon_transactions

    @classmethod
    def deduplicate_matched(cls, df: pd.DataFrame) -> pd.DataFrame:
        """
        Deduplicate Multiple Connections Made

        Parameters
        ----------
        df: pd.DataFrame

        Returns
        -------
        pd.DataFrame
        """
        deduped = df.copy()
        deduped["duplicated"] = deduped.duplicated(subset=["id"], keep=False)
        deduped = deduped[deduped["duplicated"] == False]  # noqa:E712
        return deduped

    @classmethod
    def _extract_total_from_payments(cls, df: pd.DataFrame) -> pd.DataFrame:
        """
        Extract the Credit Card Payments from the payments column

        There is quite a bit of data manipulation going on here. We
        need to extract meaningful credit card transaction info from strings like this:

        Visa ending in 9470: September 11, 2022: $29.57; \
        Visa ending in 9470: September 11, 2022: $2.22;
        """
        extracted = df.copy()
        extracted["new_total"] = extracted["payments"].str.rstrip(";").str.split(";")
        exploded_totals = extracted.explode("new_total", ignore_index=True)
        exploded_totals = exploded_totals[
            exploded_totals["new_total"].str.strip() != ""
        ]
        currency_matcher = r"(?:[\£\$\€]{1}[,\d]+.?\d*)"
        exploded_totals["parsed_total"] = exploded_totals["new_total"].str.findall(
            currency_matcher
        )
        exploded_totals = exploded_totals.explode("parsed_total", ignore_index=True)
        exploded_totals["parsed_total"] = exploded_totals["parsed_total"].str.replace(
            "[^0-9.]", "", regex=True
        )
        exploded_totals["parsed_total"] = exploded_totals["parsed_total"].astype(
            np.float64
        )
        exploded_totals = exploded_totals[~exploded_totals["parsed_total"].isnull()]
        exploded_totals["total"] = np.where(
            ~exploded_totals["parsed_total"].isnull(),
            exploded_totals["parsed_total"],
            exploded_totals["total"],
        )
        return exploded_totals

    @classmethod
    def _extract_extra_from_orders(cls, df: pd.DataFrame) -> pd.DataFrame:
        """
        Extract the Credit Card Refunds and Whole Foods Orders
        """
        refunds = df.copy()
        refunded_data = refunds[refunds["refund"] > 0].copy()
        refunded_data["total"] = -refunded_data["refund"]
        refunded_data["items"] = "REFUND: " + refunded_data["items"]
        complete_amazon_data = pd.concat([refunds, refunded_data], ignore_index=True)
        complete_amazon_data["items"] = np.where(
            complete_amazon_data["to"].str.startswith("Whole Foods"),
            "Whole Foods Groceries",
            complete_amazon_data["items"],
        )
        return complete_amazon_data

    @classmethod
    def merge_transactions(
        cls, amazon: pd.DataFrame, transactions: pd.DataFrame, time_range: int = 7
    ) -> pd.DataFrame:
        """
        Merge Amazon Transactions and LunchMoney Transaction

        Parameters
        ----------
        amazon: pd.DataFrame
        transactions: pd.DataFrame
        time_range: int
            Number of days used to connect credit card transactions with
            Amazon transactions

        Returns
        -------
        pd.DataFrame
        """
        exploded_totals = cls._extract_total_from_payments(df=amazon)
        complete_amazon_data = cls._extract_extra_from_orders(df=exploded_totals)
        merged_data = transactions.copy()
        merged_data = merged_data.merge(
            complete_amazon_data,
            how="inner",
            left_on=["amount"],
            right_on=["total"],
            suffixes=(None, "_amazon"),
        )
        merged_data["start_date"] = merged_data["date_amazon"]
        merged_data["end_date"] = merged_data["date_amazon"] + datetime.timedelta(
            days=time_range
        )
        merged_data.query(
            "start_date <= date <= end_date",
            inplace=True,
        )
        merged_data["notes"] = merged_data["items"]
        deduplicated = cls.deduplicate_matched(df=merged_data)
        logger.info("%s Matching Amazon Transactions Identified", len(deduplicated))
        return deduplicated[TransactionObject.__fields__.keys()]

    def cache_transactions(
        self, start_date: datetime.date, end_date: datetime.date
    ) -> dict[int, TransactionObject]:
        """
        Cache Transactions to Memory

        Parameters
        ----------
        start_date : datetime.date
        end_date : datetime.date

        Returns
        -------
        Dict[int, TransactionObject]
        """
        end_cache_date = end_date + datetime.timedelta(days=self.time_window)
        logger.info(
            "Fetching LunchMoney transactions between %s and %s",
            start_date,
            end_cache_date,
        )
        self.get_latest_cache(include=[CategoriesObject, UserObject])
        self.refresh_transactions(start_date=start_date, end_date=end_cache_date)
        logger.info(
            'Scanning LunchMoney Budget: "%s"',
            html.unescape(self.lunch_data.user.budget_name),
        )
        logger.info(
            "%s transactions returned from LunchMoney",
            len(self.lunch_data.transactions),
        )
        return self.lunch_data.transactions

    def print_transaction(
        self, transaction: TransactionObject, former_transaction: TransactionObject
    ) -> None:
        """
        Print a Transaction for interactive input
        """
        transaction_table = table.Table(show_header=False)
        notes_table = table.Table(show_header=False)
        transaction_table.add_row("🛒 Transaction ID", str(former_transaction.id))
        transaction_table.add_row("🏦 Payee", former_transaction.payee)
        transaction_table.add_row("📅 Date", str(former_transaction.date))
        transaction_table.add_row(
            "💰 Amount", self.format_currency(amount=former_transaction.amount)
        )
        if former_transaction.category_id is not None:
            transaction_table.add_row(
                "📊 Category",
                self.lunch_data.categories[former_transaction.category_id].name,
            )
        if (
            former_transaction.original_name is not None
            and former_transaction.original_name != former_transaction.payee
        ):
            transaction_table.add_row(
                "🏦 Original Payee", former_transaction.original_name
            )
        if former_transaction.notes is not None:
            transaction_table.add_row("📝 Notes", former_transaction.notes)
        notes_table.add_row(
            "🗒  Amazon Notes",
            transaction.notes.strip(),  # type: ignore[union-attr]
        )
        print()
        print(transaction_table)
        print(notes_table)

    def update_transaction(
        self, transaction: TransactionObject, confirm: bool = True
    ) -> Optional[dict[str, Any]]:
        """
        Update a Transaction's Notes if they've changed

        Parameters
        ----------
        transaction: TransactionObject
        confirm: bool

        Returns
        -------
        Optional[Dict[str, Any]]
        """
        former_transaction = self.lunch_data.transactions[transaction.id]
        response = None
        stripped_notes = transaction.notes.strip()  # type: ignore[union-attr]
        acceptable_length = min(349, len(stripped_notes))
        new_notes = stripped_notes[:acceptable_length]
        if former_transaction.notes != new_notes:
            confirmation = True
            if confirm is True:
                self.print_transaction(
                    transaction=transaction, former_transaction=former_transaction
                )
                confirmation = Confirm.ask(
                    f"\t❓ Should we update transaction #{transaction.id}?"
                )
            if confirmation is True:
                response = self.lunch.update_transaction(
                    transaction_id=transaction.id,
                    transaction=TransactionUpdateObject(notes=new_notes),
                )
                if confirm is True:
                    print(f"\t✅ Transaction #{transaction.id} updated")
        return response

    def process_transactions(self, confirm: bool = True) -> None:
        """
        Run the End-to-End Process
        """
        logger.info(
            "Beginning search to match Amazon and LunchMoney - using %s day window",
            self.time_window,
        )
        amazon_df = self.amazon_to_df()
        min_date = amazon_df["date"].min().to_pydatetime().date()
        max_date = amazon_df["date"].max().to_pydatetime().date()
        logger.info(
            "%s Amazon transactions loaded ranging from %s to %s",
            len(amazon_df),
            min_date,
            max_date,
        )
        self.cache_transactions(start_date=min_date, end_date=max_date)
        transaction_df = self.models_to_df(
            models=self.lunch_data.transactions.values(),
        )
        amazon_transaction_df = self.filter_amazon_transactions(df=transaction_df)
        merged_data = self.merge_transactions(
            transactions=amazon_transaction_df,
            amazon=amazon_df,
            time_range=self.time_window,
        )
        updated_transactions = self.df_to_models(
            df=merged_data, model_type=TransactionObject
        )
        responses = []
        for item in updated_transactions:
            resp = self.update_transaction(transaction=item, confirm=confirm)
            if resp is not None:
                responses.append(resp)
        logger.info("%s LunchMoney transactions updated", len(responses))

    @staticmethod
    def format_currency(amount: float) -> str:
        """
        Format currency amounts to be pleasant and human readable

        Parameters
        ----------
        amount: float
            Float Amount to be converted into a string

        Returns
        -------
        str
        """
        if amount < 0:
            float_string = f"[bold red]$ ({float(abs(amount)):,.2f})[/bold red]"
        else:
            float_string = f"[bold green]$ {float(amount):,.2f}[/bold green]"
        return float_string


@click.command("run")
@click.option(
    "-f",
    "--file",
    "csv_file",
    type=click.Path(exists=True, resolve_path=True),
    help="File Path of the Amazon Export",
    required=True,
)
@click.option(
    "-w",
    "--window",
    "window",
    type=click.INT,
    help="Allowable time window between Amazon transaction date and "
    "credit card transaction date",
    default=7,
)
@click.option(
    "-a",
    "--all",
    "update_all",
    is_flag=True,
    type=click.BOOL,
    help="Whether to skip the confirmation step and simply update all matched "
    "transactions",
    default=False,
)
@click.option(
    "-t",
    "--token",
    "access_token",
    type=click.STRING,
    help="LunchMoney Access Token - defaults to the LUNCHMONEY_ACCESS_TOKEN environment variable",
    envvar="LUNCHMONEY_ACCESS_TOKEN",
)
def run_primelunch(
    csv_file: str, window: int, update_all: bool, access_token: str
) -> None:
    """
    Run the PrimeLunch Update Process
    """
    primelunch = PrimeLunch(
        file_path=csv_file, time_window=window, access_token=access_token
    )
    primelunch.process_transactions(confirm=not update_all)


if __name__ == "__main__":
    set_up_logging(log_level=logging.INFO)
    run_primelunch()



"""
Base App Class
"""

import base64
import datetime
import json
import logging
import shutil
from abc import ABC, abstractmethod
from hashlib import sha256
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union

from pydantic import BaseModel
from pydantic.json import pydantic_encoder

from lunchable import LunchMoney
from lunchable._config import FileConfig
from lunchable.models import (
    AssetsObject,
    CategoriesObject,
    CryptoObject,
    LunchableModel,
    PlaidAccountObject,
    TagsObject,
    TransactionObject,
    UserObject,
)

logger = logging.getLogger(__name__)

LunchableModelType = TypeVar("LunchableModelType", bound=LunchableModel)


class LunchableDataModel(LunchableModel):
    """
    Core Data Model Defining App Dependencies
    """

    model: Type[LunchableModel]
    function: Callable[[Any], Any]
    kwargs: Dict[str, Any] = {}


class LunchableDataContainer(BaseModel):
    """
    Data Container for Lunchable App Data
    """

    plaid_accounts: Dict[int, PlaidAccountObject] = {}
    transactions: Dict[int, TransactionObject] = {}
    categories: Dict[int, CategoriesObject] = {}
    assets: Dict[int, AssetsObject] = {}
    tags: Dict[int, TagsObject] = {}
    user: UserObject = UserObject(
        user_id=0, user_name="", user_email="", account_id=0, budget_name=""
    )
    crypto: Dict[int, CryptoObject] = {}

    @property
    def asset_map(self) -> Dict[int, Union[PlaidAccountObject, AssetsObject]]:
        """
        Asset Mapping Across Plaid Accounts and Assets

        Returns
        -------
        Dict[int, Union[PlaidAccountObject, AssetsObject]]
        """
        asset_map: Dict[int, Union[PlaidAccountObject, AssetsObject]] = {}
        asset_map.update(self.plaid_accounts)
        asset_map.update(self.assets)
        return asset_map

    @property
    def plaid_accounts_list(self) -> List[PlaidAccountObject]:
        """
        List of Plaid Accounts

        Returns
        -------
        List[PlaidAccountObject]
        """
        return list(self.plaid_accounts.values())

    @property
    def assets_list(self) -> List[AssetsObject]:
        """
        List of Assets

        Returns
        -------
        List[AssetsObject]
        """
        return list(self.assets.values())

    @property
    def transactions_list(self) -> List[TransactionObject]:
        """
        List of Transactions

        Returns
        -------
        List[TransactionObject]
        """
        return list(self.transactions.values())

    @property
    def categories_list(self) -> List[CategoriesObject]:
        """
        List of Categories

        Returns
        -------
        List[CategoriesObject]
        """
        return list(self.categories.values())

    @property
    def tags_list(self) -> List[TagsObject]:
        """
        List of Tags

        Returns
        -------
        List[TagsObject]
        """
        return list(self.tags.values())

    @property
    def crypto_list(self) -> List[CryptoObject]:
        """
        List of Crypto

        Returns
        -------
        List[CryptoObject]
        """
        return list(self.crypto.values())


class BaseLunchableApp(ABC):
    """
    Abstract Base Class for Lunchable Apps
    """

    __lunchable_object_mapping__: Dict[str, str] = {
        PlaidAccountObject.__name__: "plaid_accounts",
        TransactionObject.__name__: "transactions",
        CategoriesObject.__name__: "categories",
        AssetsObject.__name__: "assets",
        TagsObject.__name__: "tags",
        UserObject.__name__: "user",
        CryptoObject.__name__: "crypto",
    }

    @property
    @abstractmethod
    def lunchable_models(self) -> List[LunchableDataModel]:
        """
        Every LunchableApp should define which data objects it depends on

        Returns
        -------
        List[LunchableDataModel]
        """

    @property
    @abstractmethod
    def __builtin_data_models__(self) -> List[LunchableDataModel]:
        """
        Every LunchableApp should define which data objects are built-in

        Returns
        -------
        List[LunchableDataModel]
        """

    def __init__(self, cache_time: int = 0, access_token: Optional[str] = None):
        """
        Lunchable App Initialization

        Parameters
        ----------
        cache_time: int
            Amount of time until the cache should be refreshed
            (in seconds). Defaults to 0 which always polls for the latest data
        access_token: Optional[str]
            Lunchmoney Developer API Access Token
        """
        self.lunch = LunchMoney(access_token=access_token)
        self.lunch_data = LunchableDataContainer()
        self.data_dir = FileConfig.DATA_DIR.joinpath(self.__class__.__name__).joinpath(
            sha256(self.lunch.access_token.encode("utf-8")).hexdigest()
        )
        self.cache_time = cache_time
        if self.cache_time > 0:
            self.data_dir.mkdir(exist_ok=True, parents=True)

    def _cache_single_object(
        self,
        model: Type[LunchableModelType],
        function: Callable[[Any], Any],
        kwargs: Optional[Dict[str, Any]] = None,
        force: bool = False,
    ) -> Union[LunchableModelType, List[LunchableModelType]]:
        """
        Cache a Core Lunchable Data Object

        Parameters
        ----------
        model: Type[LunchableModel]
        function: Callable
        kwargs: Optional[Dict[str, Any]]
        force: bool

        Returns
        -------
        Any
        """
        if kwargs is None:
            kwargs = {}
        data_file = self.data_dir.joinpath(f"{model.__name__}.lunch")
        if force is True:
            refresh = True
        elif self.cache_time > 0 and data_file.exists():
            modified_time = datetime.datetime.fromtimestamp(
                data_file.stat().st_mtime, tz=datetime.timezone.utc
            )
            current_time = datetime.datetime.now(tz=datetime.timezone.utc)
            file_age = current_time - modified_time
            refresh = file_age > datetime.timedelta(seconds=self.cache_time)
        else:
            refresh = True
        if refresh is True:
            data_objects = function(**kwargs)  # type: ignore[call-arg]
            if self.cache_time > 0:
                plain_data: str = json.dumps(data_objects, default=pydantic_encoder)
                base64_data: bytes = base64.b64encode(plain_data.encode("utf-8"))
                data_file.write_bytes(base64_data)
        else:
            file_text: bytes = data_file.read_bytes()
            json_body: bytes = base64.b64decode(file_text)
            json_data: Union[Dict[str, Any], List[Dict[str, Any]]] = json.loads(
                json_body.decode("utf-8")
            )
            if isinstance(json_data, dict):
                data_objects = model(**json_data)
            else:
                data_objects = [model(**item) for item in json_data]
        return data_objects

    def get_latest_cache(
        self,
        include: Optional[List[Type[LunchableModel]]] = None,
        exclude: Optional[List[Type[LunchableModel]]] = None,
        force: bool = False,
    ) -> None:
        """
        Cache the Underlying Data Objects

        Parameters
        ----------
        include : Optional[List[Type[LunchableModel]]]
            Models to refresh cache for (instead of all of them)
        exclude : Optional[List[Type[LunchableModel]]]
            Models to skip cache refreshing
        force: bool
            Whether to force the cache

        Returns
        -------
        None
        """
        models_to_process = self.lunchable_models + self.__builtin_data_models__
        if include is not None:
            new_models_to_process: List[LunchableDataModel] = []
            data_model_mapping = {item.model: item for item in models_to_process}
            for model_class in include:
                new_models_to_process.append(data_model_mapping[model_class])
            models_to_process = new_models_to_process
        exclusions = exclude if exclude is not None else []
        for data_model in models_to_process:
            if data_model.model in exclusions:
                continue
            cache = self._cache_single_object(
                model=data_model.model,
                function=data_model.function,
                kwargs=data_model.kwargs,
                force=force,
            )
            cache_attribute: Union[Dict[int, LunchableModel], LunchableModel]
            if isinstance(cache, list):
                cache_attribute = {item.id: item for item in cache}
            else:
                cache_attribute = cache
            setattr(
                self.lunch_data,
                self.__lunchable_object_mapping__[data_model.model.__name__],
                cache_attribute,
            )

    def refresh_transactions(
        self,
        start_date: Optional[Union[datetime.date, datetime.datetime, str]] = None,
        end_date: Optional[Union[datetime.date, datetime.datetime, str]] = None,
        tag_id: Optional[int] = None,
        recurring_id: Optional[int] = None,
        plaid_account_id: Optional[int] = None,
        category_id: Optional[int] = None,
        asset_id: Optional[int] = None,
        group_id: Optional[int] = None,
        is_group: Optional[bool] = None,
        status: Optional[str] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        debit_as_negative: Optional[bool] = None,
        pending: Optional[bool] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Dict[int, TransactionObject]:
        """
        Refresh App data with the latest transactions

        start_date: Optional[Union[datetime.date, datetime.datetime, str]]
            Denotes the beginning of the time period to fetch transactions for. Defaults
            to beginning of current month. Required if end_date exists. Format: YYYY-MM-DD.
        end_date: Optional[Union[datetime.date, datetime.datetime, str]]
            Denotes the end of the time period you'd like to get transactions for.
            Defaults to end of current month. Required if start_date exists.
        tag_id: Optional[int]
            Filter by tag. Only accepts IDs, not names.
        recurring_id: Optional[int]
            Filter by recurring expense
        plaid_account_id: Optional[int]
            Filter by Plaid account
        category_id: Optional[int]
            Filter by category. Will also match category groups.
        asset_id: Optional[int]
            Filter by asset
        group_id: Optional[int]
            Filter by group_id (if the transaction is part of a specific group)
        is_group: Optional[bool]
            Filter by group (returns transaction groups)
        status: Optional[str]
            Filter by status (Can be cleared or uncleared. For recurring
            transactions, use recurring)
        offset: Optional[int]
            Sets the offset for the records returned
        limit: Optional[int]
            Sets the maximum number of records to return. Note: The server will not
            respond with any indication that there are more records to be returned.
            Please check the response length to determine if you should make another
            call with an offset to fetch more transactions.
        debit_as_negative: Optional[bool]
            Pass in true if you'd like expenses to be returned as negative amounts and
            credits as positive amounts. Defaults to false.
        pending: Optional[bool]
            Pass in true if you'd like to include imported transactions with a pending status.
        params: Optional[dict]
            Additional Query String Params

        Returns
        -------
        Dict[int, TransactionObject]
        """
        transactions = self.lunch.get_transactions(
            start_date=start_date, end_date=end_date, status=status
        )
        transaction_map = {item.id: item for item in transactions}
        self.lunch_data.transactions = transaction_map
        return transaction_map

    def delete_cache(self) -> None:
        """
        Delete any corresponding cache files
        """
        if self.data_dir.exists():
            shutil.rmtree(self.data_dir)


class LunchableApp(BaseLunchableApp):
    """
    Pre-Built Lunchable App

    This app comes with a `data` property which represents all the base data
    the app should need. Extend the `data_models` property to items like
    `TransactionObject`s to interact with transactions
    """

    @property
    def lunchable_models(self) -> List[LunchableDataModel]:
        """
        Which Data Should this app get
        """
        return []

    @property
    def __builtin_data_models__(self) -> List[LunchableDataModel]:
        """
        Built-In Models to Populate Most LunchableApp instances

        Returns
        -------
        List[LunchableDataModel]
        """
        return [
            LunchableDataModel(
                model=CategoriesObject, function=self.lunch.get_categories
            ),
            LunchableDataModel(
                model=PlaidAccountObject,
                function=self.lunch.get_plaid_accounts,
            ),
            LunchableDataModel(
                model=AssetsObject,
                function=self.lunch.get_assets,
            ),
            LunchableDataModel(
                model=TagsObject,
                function=self.lunch.get_tags,
            ),
            LunchableDataModel(
                model=UserObject,
                function=self.lunch.get_user,
            ),
            LunchableDataModel(
                model=CryptoObject,
                function=self.lunch.get_crypto,
            ),
        ]


class LunchableTransactionsBaseApp(LunchableApp, ABC):
    """
    LunchableApp supporting transactions
    """

    data_models: List[LunchableDataModel] = []

    @property
    @abstractmethod
    def start_date(self) -> datetime.date:
        """
        LunchableTransactionsApp requires a Start Date

        Returns
        -------
        datetime.date
        """

    @property
    @abstractmethod
    def end_date(self) -> datetime.date:
        """
        LunchableTransactionsApp requires a End Date

        Returns
        -------
        datetime.date
        """

    @property
    def __builtin_data_models__(self) -> List[LunchableDataModel]:
        """
        Which Data Should this app get
        """
        return [
            *super().__builtin_data_models__,
            LunchableDataModel(
                model=TransactionObject,
                function=self.lunch.get_transactions,
                kwargs={"start_date": self.start_date, "end_date": self.end_date},
            ),
        ]

    def refresh_transactions(  # type: ignore[override]
        self,
        start_date: Optional[datetime.date] = None,
        end_date: Optional[datetime.date] = None,
    ) -> Dict[int, TransactionObject]:
        """
        Refresh App data with the latest transactions

        Returns
        -------
        Dict[int, TransactionObject]
        """
        transactions = self.lunch.get_transactions(
            start_date=start_date if start_date is not None else self.start_date,
            end_date=end_date if end_date is not None else self.end_date,
        )
        transaction_map = {item.id: item for item in transactions}
        self.lunch_data.transactions = transaction_map
        return transaction_map


class LunchableTransactionsApp(LunchableTransactionsBaseApp):
    """
    Pre-Built Lunchable App with the last 365 days worth of transactions
    """

    @property
    def start_date(self) -> datetime.date:
        """
        LunchableTransactionsApp requires a Start Date

        Returns
        -------
        datetime.date
        """
        today = datetime.date.today()
        return today.replace(year=today.year - 100)

    @property
    def end_date(self) -> datetime.date:
        """
        LunchableTransactionsApp requires a End Date

        Returns
        -------
        datetime.date
        """
        today = datetime.date.today()
        return today.replace(year=today.year + 100)



"""
Run Tests on the Pushover Plugin
"""

import logging
from typing import List

from lunchable.models import TransactionObject
from lunchable.plugins.pushlunch import PushLunch
from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@lunchable_cassette
def test_send_notification():
    """
    Send a Generic Notification
    """
    pusher = PushLunch()
    pusher.send_notification(
        message="This is a test notification from lunchable.", title="Test"
    )


@lunchable_cassette
def test_post_transaction(test_transactions: List[TransactionObject]):
    """
    Send
    """
    pusher = PushLunch()
    example_notification = test_transactions[0]
    example_notification.payee = "Test"
    example_notification.notes = "Example Test Notification from lunchable"
    pusher.post_transaction(transaction=example_notification)



"""
Sphinx Documentation Generator
"""

import sys
from datetime import datetime
from pathlib import Path

_project_path = Path(__file__).resolve().parent.parent.parent
_project_dir = str(_project_path)
sys.path.insert(0, _project_dir)

from lunchable._version import __application__, __author__, __version__  # noqa

_author = __author__
project = __application__
copyright = f"{datetime.now().year}, {_author}"
author = _author
release = version = __version__

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.apidoc",
    "sphinxcontrib.autodoc_pydantic",
    "autodocsumm",
    "myst_parser",
    "autoclasstoc",
    "sphinx_copybutton",
    "sphinx_autodoc_typehints",
    "sphinx_autodoc_defaultargs",
    "sphinx_click",
]

myst_heading_anchors = 5
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

templates_path = ["_templates"]
html_static_path = ["_static"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

html_theme = "sphinx_rtd_theme"
html_theme_options = {
    "display_version": True,
}
html_show_sphinx = False
html_show_sourcelink = False
# html_logo = "_static/lunchable.png"
html_favicon = "https://juftin.com/favicon.ico"

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}

autosummary_generate = True
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

source_code_dir = _project_path.joinpath("lunchable")
apidoc_module_dir = str(source_code_dir)
apidoc_output_dir = str(source_code_dir.parent.joinpath("docs", "source", "api"))
apidoc_excluded_paths = ["tests"]
apidoc_separate_modules = True

autodoc_member_order = "bysource"
autodoc_pydantic_model_show_json = False
autodoc_pydantic_settings_show_json = False

always_document_default_args = True
docstring_default_arg_substitution = "**[Default]:**"

autodoc_default_options = {"exclude-members": "with_traceback"}

html_favicon = "https://juftin.com/favicon.ico"
html_context = {
    "display_github": True,
    "github_user": "juftin",
    "github_repo": "lunchable",
    "github_version": "main/docs/source/",
}



"""
Base Apps for Plugins
"""

from .base_app import LunchableApp, LunchableModelType, LunchableTransactionsApp

__all__ = [
    "LunchableTransactionsApp",
    "LunchableApp",
    "LunchableModelType",
]



"""
Apps with Pandas Support
"""

from typing import Iterable, List, Type

import numpy as np
import pandas as pd

from lunchable.models import LunchableModel
from lunchable.plugins import LunchableApp, LunchableModelType, LunchableTransactionsApp


class LunchablePandasApp(LunchableApp):
    """
    LunchableApp with Pandas Super Powers
    """

    @staticmethod
    def models_to_df(models: Iterable[LunchableModel]) -> pd.DataFrame:
        """
        Convert Transactions Array to DataFrame

        Parameters
        ----------
        models: List[LunchableModel]

        Returns
        -------
        pd.DataFrame
        """
        if not isinstance(models, list):
            models = list(models)
        return pd.DataFrame(
            [item.model_dump() for item in models],
            columns=models[0].__fields__.keys(),
        )

    @staticmethod
    def df_to_models(
        df: pd.DataFrame, model_type: Type[LunchableModelType]
    ) -> List[LunchableModelType]:
        """
        Convert DataFrame to Transaction Array

        Parameters
        ----------
        df: pd.DataFrame
        model_type: Type[LunchableModel]

        Returns
        -------
        List[LunchableModel]
        """
        array_df = df.copy()
        array_df = array_df.fillna(np.NaN).replace([np.NaN], [None])
        model_array = array_df.to_dict(orient="records")
        return [model_type(**item) for item in model_array]


class LunchablePandasTransactionsApp(LunchableTransactionsApp, LunchablePandasApp):
    """
    LunchableTransactionsApp with Pandas Super Powers
    """



"""
Run Tests on the Splitwise Plugin
"""

import json
import logging
from os import path

import pytest

from tests.conftest import lunchable_cassette

logger = logging.getLogger(__name__)


@pytest.mark.filterwarnings("ignore:datetime.datetime.utcfromtimestamp")
def test_import_splitwise():
    """
    Try to import splitwise and succeed since the extra is installed in tox
    """
    test_case = True
    try:
        from lunchable.plugins.splitlunch import SplitLunch  # noqa
    except ImportError:
        test_case = False
    finally:
        assert test_case is True


@pytest.mark.filterwarnings("ignore:datetime.datetime.utcfromtimestamp")
@lunchable_cassette
def test_update_balance():
    """
    Update the Balance
    """
    from lunchable.plugins.splitlunch import SplitLunch

    lunch = SplitLunch()
    lunch.update_splitwise_balance()


@pytest.mark.filterwarnings("ignore:datetime.datetime.utcfromtimestamp")
def test_financial_impact():
    """
    Test the financial impact algorithm
    """
    import splitwise

    from lunchable.plugins.splitlunch.lunchmoney_splitwise import _get_splitwise_impact

    for [file, expected_self_paid, expected_impact] in [
        # For both expenses and transfers, when someone else pays, financial impact should be
        # positive
        ["splitwise_non_user_paid_expense.json", False, 9.99],
        ["splitwise_non_user_paid_transfer.json", False, 523.84],
        # When you pay, financial impact should be negative
        ["splitwise_user_paid_expense.json", True, -61.65],
        ["splitwise_user_paid_transfer.json", True, -431.92],
        # And any transaction that doesn't involve you should have no impact
        ["splitwise_non_involved_expense.json", False, 0],
        ["splitwise_non_involved_transfer.json", False, 0],
    ]:
        with open(path.join(path.dirname(__file__), f"data/{file}")) as json_file:
            expense = splitwise.Expense(json.load(json_file))
        financial_impact, self_paid = _get_splitwise_impact(
            expense=expense, current_user_id=1234059
        )
        assert (
            self_paid is expected_self_paid
        ), f"Expected {expected_self_paid} for {file}"
        assert (
            financial_impact == expected_impact
        ), f"Expected {expected_impact} for {file}"
